#ifndef jvmComm
#define jvmComm

/*
*****************************************************
* jvmComm - JVM communication with programming port *
*****************************************************
*/

#include <propeller.h>
#include <string.h>

//Javelin IDE constants
#define  kReplyBit           0x80
#define  kFlag               0x7E
#define  kEscape             0x7D
#define  kInterrupt          0x19
#define  kResetWait          10000 // Time taken for SX to reset.
#define  kAckWait            5000  // The amount of time (in milliseconds) to wait for acknowledgement.
#define  kResetCharWait      3     // The amount of time to wait before sending first char after a reset.
#define  kProgramWait        2000  // The amount of time (in milliseconds) to wait for acknowledgement when programming.
#define  kStepWait           20000 // The amount of time (in milliseconds) to wait for acknowledgement when stepping.
#define  kBaseVersion        0x50  // Compatible base version of the firmware.
#define  kMaxQueueLength     500   // Maximum number of characters than can be queued.
#define  kComNone            0
#define  kComStatus          1
#define  kComReset           2
#define  kComStep            3
#define  kComRun             4
#define  kComError           5
#define  kComStop            6
#define  kComStack           7
#define  kComConfig          8
#define  kComHeap            9
#define  kComBreakpoints     10
#define  kComProgram         11
#define  kComDownloadOK      12
#define  kComStepDone        13
#define  kComAtn             14
#define  kComStartLoad       15
//#define  kComEndLoad         kBaseVersion
#define  kComTerminalQuery   17     // Check for data from the terminal.
#define  kComTerminalData    18     // Data follows.
#define  kComTerminalNoData  19     // No data follows.
#define  kComDownloadRamOK   20

//values following kComError
#define  kErrNone        0
#define  kErrOpcode      1
#define  kErrBreakpoint  2
#define  kErrMessage     3   //to send message to IDE message window
#define  kErrException   4   //to send exception to IDE status window
#define  kErrEnd         5   //to signal end of java program
#define  kErrJvm         6   //to send message to IDE status window


//Spin stamp debug pin assignments
#define  stampSOUT  16 //serial out (pin 1 of spin stamp)
#define  stampSIN   17 //serial in  (pin 2 of spin stamp)
#define  stampATN   18 //digital in (pin 3 of spin stamp, when activated, do reboot)

//Propeller system pin assignments
#define  propSCL  28 //external eeprom SCL
#define  propSDA  29 //external eeprom SDA
#define  propTX   30 //programming output
#define  propRX   31 //programming input
  
#include "jvmSerial.c"
#include "jvmData.c"
  
  
long atnStack[10];     //stack for monitoring ATN pin
int  jemArrayPtr;      //for downloading jembytes into javaProg array
int  rangestart;  
int  rangelength; 
char packet[32];       //buffer to hold packets
char reset;       
char idePort;     
char loadRam;          //set to true if downloaded jembytes only to ram

  
static void atnReset(void) {
  while (true) { //keep running
    if (reset == false) {
      while ((INA & (1<<stampATN)) == 0) ; //wait for rising edge  
      //stop();
      while ((INA & (1<<stampATN)) != 0) ; //wait for falling edge
      //switch (idePort) {
      //  case 1: start(propRX,propTX,0,28800);
      //          break;
	  //  case 2: start(stampSIN,stampSOUT,0,28800);
      //          break;
	  //  default:
	  //}
	  //start(stampSIN,stampSOUT,0,28800);
      reset = true;                    //now enter reset state
    }
  }
}

static void sendByte(long txbyte) {
  if ((txbyte == kEscape) || (txbyte == kFlag)) {
    txbyte ^= 0x20;
    tx(kEscape);
  }
  tx(txbyte);
}
  
static void sendFlag(void) {
/* send kFlag value */
  tx(kFlag);
}

static long ComputeChecksum(long packetAddr, long len) {
  long checksum;
/* compute checksum for packet */
  checksum = 0;
  while (len-- > 0)
    checksum += *(char *)packetAddr++;
  return (-checksum) & 0xFF;
}

static long CheckChecksum(long packetAddr, long len) {
  long checksum;
/* check checksum for packet */
  checksum = 0;
  while (len-- > 0)
    checksum += *(char *)packetAddr++;
  return (checksum == 0);
}

static void sendPacket(long str, long len) {
  char *s;
/* send packet */
  s = (char *)str;
  if (len > 0) {
    while (len > 0) {
      sendByte(*s++);
      len--;
    }
	sendFlag();  
  }
}  

void comStart(long port, long monitor) {
/* start serial driver, must only be called from main */
  idePort = port;
  switch (port) {
    case 1: start(propRX,propTX,0,28800);
	        break;
    case 2: start(stampSIN,stampSOUT,0,28800);
	        break;
    //default:
  }
  //monitor ATN pin (optional) to simulate reset
  if (monitor && (port == 2))
    cognew(atnReset,atnStack);
}

void doReset(void) {
/* send reset sequence */
  sendByte(kComStartLoad);   //send 0x0F
  wait(CLKFREQ/20 + CNT); //wait 0.05 second
  sendByte(getEndLoad());   //send device version (0x50, 0x70, 0x78)
  wait(CLKFREQ/20 + CNT); //wait 0.05 second
}

long doDebug(long timeout) {
  long value,c,i,j,chk;
/* poll communication port for debug commands */
  if (timeout == 0)
    c = rxcheck();
  else
    c = rxtime(timeout);
  switch (c) {
    case -1: return -1;
    
    case kComStatus: //IDE sent QueryStatus command (bytes kComStatus, )
                     //IDE expects packet of 6 bytes
                     //(kComStatus|kReplyBit),jvm_pc_lsb,jvm_pc_msb,activity,bytecode,checksum
                     sendFlag(); //resetflag
                     packet[0] = kComStatus | kReplyBit;
                     i = getJvmPC()+1; //IDE subtracts 1 to compensate for the increment in the JVM
                     packet[1] = i; //pc_lsb
                     packet[2] = i>>8; //pc_msb
                     packet[3] = getActivity() & 0x7F;  //after reset activity = kPassive (=0)
                     packet[4] = readJavaByte(getJvmPC());
                     packet[5] = ComputeChecksum((long)packet,5);
                     sendPacket((long)packet,6);
                     return 0;
                      
    case kComConfig: //IDE sent QueryConfig command (bytes kComConfig, )
                     //IDE expects packet of 8 bytes
                     //(kComConfig|kReplyBit),memsize_lsb,memsize_msb,progsize_lsb,progsize_msb,heapbase_lsb,heapbase_msb,checksum
                     sendFlag(); //resetflag
                     packet[0] = kComConfig | kReplyBit;
                     packet[1] = getMemSize();
                     packet[2] = getMemSize() >> 8;
                     packet[3] = getProgSize();
                     packet[4] = getProgSize() >> 8;
                     packet[5] = getHeapBase();
                     packet[6] = getHeapBase() >> 8;
                     packet[7] = ComputeChecksum((long)packet,7);
                     sendPacket((long)packet,8);
                     return 0;

    case kComStep:   //IDE sent Step command
                     //respond with (kComStep|kReplyBit), do step, send (kComStepDone|kReplyBit)
                     sendByte(kComStep | kReplyBit);
                     setNotifyStepDone(); //notify jvm to signal IDE after step done
                     setMustDoStep();     //notify jvm to perform a step
                     return 0;
                    
    case kComRun:    //IDE sent Run command
                     //run java program at full speed, respond with (kComRun|kReplyBit)
                     sendByte(kComRun | kReplyBit);
                     setActivity(kRun);
                     return 0;
                    
    case kComStop:   //IDE sent Stop command
                     //stop java program, respond with kComStop|kReplyBit
                     setActivity(kPassive);
                     sendByte(kComStop | kReplyBit);
                     return 0;
                    
    case kComStack:  //IDE sent QueryStack command
                     //IDE expects packet of variable length
                     //(kComStack|kReplyBit),jvmSP_lsb,jvmSP_msb,jvmFP_lsb,jvmFP_msb,jvmMP_lsb,jvmMP_msb,stackbytes,checksum
                     sendFlag(); //resetflag
                     chk = 0;
                     sendByte(kComStack | kReplyBit);
                     chk += (kComStack | kReplyBit);

                     sendByte(getJvmSP()); //sp_lsb
                     chk += getJvmSP();
                     sendByte(getJvmSP() >> 8); //sp_msb
                     chk += (getJvmSP() >> 8);

                     sendByte(getJvmFP()); //fp_lsb
                     chk += getJvmFP();
                     sendByte(getJvmFP() >> 8); //fp_msb
                     chk += (getJvmFP() >> 8);

                     sendByte(getJvmMP()); //mp_lsb
                     chk += getJvmMP();
                     sendByte(getJvmMP() >> 8); //mp_msb
                     chk += (getJvmMP() >> 8);

                     //send stack bytes
                     i = getJvmSB();
                     do {
                       j = readJavaByte(i) & 0xFF; //stack data
                       sendByte(j);
                       chk += j;
                       i++;
                     } while (i != getJvmSP());

                     chk = (chk ^ 0xFF) + 1;
                     chk &= 0xFF;
                     sendByte(chk);  //checksum
                     sendFlag(); //endflag  
                     return 0;
                    
    case kComHeap:   //IDE sent QueryHeap command
                     //IDE sends 5 byte packet
                     //kComHeap,rangestart_msb,rangestart_lsb,rangelength_msb,rangelength_lsb
                     //IDE expects packet of rangelength+4 bytes
                     //(kComHeap|kReplyBit),heapaddr_msb,heapaddr_lsb,<rangelength bytes>,checksum
                     //get packet
                     packet[0] = c;
                     i = 1;
                     while (i < 5) {
                       c = rxtime(1000);
                       if ((c == kEscape) || (c == kFlag)) {
                         c = rxtime(100);
                         c ^= 0x20;
                       }
					   packet[i++] = c;
                     }
					 sendFlag(); //resetflag
                     rangestart = packet[1] << 8;
                     rangestart += packet[2];
                     rangelength = packet[3] << 8;
                     rangelength += packet[4];
                     chk = 0;
                     sendByte(kComHeap | kReplyBit);
                     chk += (kComHeap | kReplyBit);

                     sendByte(getJvmHP() >> 8); //hp_msb
                     chk += (getJvmHP() >> 8);
                     sendByte(getJvmHP()); //hp_lsb
                     chk += getJvmHP();

                     //send requested bytes
                     while (rangelength > 0) {
                       i = readJavaByte(rangestart);
                       sendByte(i);
                       chk += i;
                       rangestart++;
                       if (rangestart == getMemSize())
                         rangestart = 0;
                       rangelength--;
                     }
					 
                     chk = (chk ^ 0xFF) + 1;
                     chk &= 0xFF;
                     sendByte(chk);  //checksum
                     sendFlag(); //endflag  
                     return 0;
                    
    case kComBreakpoints: //IDE sent DownloadBreakpoints command
                          //IDE sends variable length packet
                          //packet[0] = kComBreakpoints
                          //packet[1] = count (number of breakpoints)
                          //count words breakpoint_lsb,breakpoint_msb (last BP word can be BP for step)
                          //checksum
                          //Handle list of breakpoints and respond with kComBreakpoints|kReplyBit
                          //get packet
                          packet[0] = c;
                          chk = c;
                          c = rxtime(1000);
                          if ((c == kEscape) || (c == kFlag)) {
                            c = rxtime(100);
                            c ^= 0x20;
                          }
						  chk += c;
                          setBreakpoints(c); //number of breakpoints
                          i = c + c + 1; //number of bytes to receive (c words + checksum)
                          j = 0;
                          while (i > 0) {
                            c = rxtime(1000);
                            if ((c == kEscape) || (c == kFlag)) {
                              c = rxtime(100);
                              c ^= 0x20;
                            }
							chk += c;
                            setBreakpointByte(j++,c); //breakpoints are received lsb first
                            i--;
						  }
						  if ((chk & 0xFF) != 0) {
                            wait(CLKFREQ/200 + CNT); //wait 0.005 second
                            sendByte(kComBreakpoints); //negative response
                          }
						  else {
                            wait(CLKFREQ/200 + CNT); //wait 0.005 second
                            sendByte(kComBreakpoints | kReplyBit); //positive response
                          }
						  return 0;
                         
    case kComDownloadOK:  //IDE sent DownloadOK command
                          //if ready to receive a download, respond with kComDownloadOK|kReplyBit
                          sendByte(kComDownloadOK | kReplyBit);
                          loadRam = false;
                          jemArrayPtr = 0;
                          return 0;

    case kComDownloadRamOK: //IDE sent DownloadRamOK command
                            //if ready to receive a download, respond with kComDownloadRamOK|kReplyBit
                            sendByte(kComDownloadRamOK | kReplyBit);
                            loadRam = true;
                            jemArrayPtr = 0;
                            return 0;

    case kComAtn:         //IDE sent ATN command (used to check for echoes)
                          //do nothing
                          return 0;
                         
    case kComProgram: //IDE sent kComProgram command
                      //IDE sends 19 byte packets
                      //packet[0] = kComProgram
                      //packet[1] = 0x01 for first packet, 0x00 for following packets
                      //packet[2..17] = next 16 bytes from jemfile
                      //packet[18] = checksum
                      //If checksum ok, write packet[2..17] to eeprom and respond with kComProgram|kReplyBit
                      //This state is only left when the IDE resets this device (after downloading all packets)
                      //start downloading
                      packet[0] = c;
                      //get one packet
                      i = 1;
                      chk = c;
                      while (i < 19) {
                        c = rxtime(1000);
                        if ((c == kEscape) || (c == kFlag)) {
                          c = rxtime(100);
                          c ^= 0x20;
                        }
						packet[i++] = c;
                        chk += c;
                      }
					  if ((chk & 0xFF) != 0) {
                        wait(CLKFREQ/200 + CNT); //wait 0.005 second
                        sendByte(kComProgram); //negative response
                      }
					  else {
                        if (packet[1] == 1)
                          jemArrayPtr = 0;
                        for (i=2; i<18; i++) {
                          writeJavaByte(jemArrayPtr,packet[i]);
                          jemArrayPtr++;
                        }
						if (!loadRam)
                          eeBlockWrite(jemArrayPtr-16);
                        wait(CLKFREQ/200 + CNT); //wait 0.005 second
                        sendByte(kComProgram | kReplyBit); //positive response
                      }
					  return 0;

    case kComReset: //fake reset
                    sendByte(kComReset | kReplyBit);
                    wait(CLKFREQ/200 + CNT); //wait 0.005 second
                    reset = 1;
                    return -1;
               
    //default: //unlisted value
  }         
  return 0;
}
               
void sendStepDone(void) {
  sendByte(kComStepDone | kReplyBit);
  clearNotifyStepDone();
  clearMustDoStep();
}  

void breakpoint(void) {
/* signal IDE that jvm reached a breakpoint */
/* after this, the jvm enters state kBreak */
  setActivity(kBreak);
  packet[0] = kComError;
  packet[1] = kErrBreakpoint;
  sendPacket((long)packet,2);
}   

void end(void) {
/* signal IDE that java program has ended */
/* after this, the jvm enters state kPassive */
  setActivity(kPassive);
  packet[0] = kComError;
  packet[1] = kErrEnd;
  sendPacket((long)packet,2);
}
  

void errOpcode(long jembyte) {
/* send unsupported/unimplemented jembyte to IDE */
/* after this, the jvm enters state kPassive */
  setActivity(kPassive);
  packet[0] = kComError;
  packet[1] = kErrOpcode;
  packet[2] = jembyte;
  sendPacket((long)packet,3);
}
  
void errException(long classOffset, long pc) {
/* send exception to IDE */
/* after this, the jvm enters state kPassive */
/* pc must point to the offending jemcode instruction */
  setActivity(kPassive);
  pc++;  //add 1 because IDE subtracts one
  packet[0] = kComError;
  packet[1] = kErrException;
  packet[2] = classOffset >> 8;
  packet[3] = classOffset;
  packet[4] = pc >> 8;
  packet[5] = pc;
  sendPacket((long)packet,6);
}

void status(long str) {
  long len;
/* send message to IDE status window */
/* use this to send jvm debug messages */
  len = strlen((char *)str);
  if (len > 0) {
    tx(kComError);
    tx(kErrJvm);
    sendPacket(str,len);
  }
}

void message(long str, long len) {
  long c,i;
/* send message to IDE message window */
/* this is the native CPU.message() and should not be used for debugging the jvm */
  if (len > 0) {
    i = 0;
    tx(kComError);
    tx(kErrMessage);
    while (len > 0) {
      c = readJavaByte(str + 4 + i);
      sendByte(c);
      i++;
      len--;
    }
	sendFlag();
  }  
}

/*
  What is the protocol used to send messages to and from the Sin and Sout pins?
  Debug messages sent from the Javelin conform to the following pattern: 

  0x05 0x03 <messagedata> 0x7E 

  <messagedata> is the sequence of bytes being sent by the CPU.message() native method. It is byte-stuffed.
  This means that if <messagedata> contains a byte <b> that has the value 0x7D or 0x7E then that byte is
  replaced with 0x7D <b>^0x20. i.e. 0x7D followed by the orginal byte xored with 0x20. 

  Terminal characters are sent to the Javelin using the following algorithm: 

  When it is ready to receive a byte (the Terminal.getByte() native method is called) the Javelin transmits
  0x11 out Sout. 
  If the terminal has no data to send to the Javelin it replies with 0x13. 
  If the terminal has a byte <b> to send, then it replies with 0x12 <b>. 
  After the Javelin is reset it send 0x0F followed a pause followed by 0x50 (the version of the Javelin firmware).

  Running the following simple program:

  public class HelloWorld { 
    public static void main() {
      System.out.println("Hello World"); 
    } 
  }

  will generate the following on Sout:

  0F 50 05 03 48 65 6C 6C 6F 20 57 6F 72 6C 64 7E 05 03 0D 0A 7E 05 05 7E 

  Note that the sequence 0x05 0x05 0x7E indicates that the Java program has ended.

*/

  
long getByte(void) {
  long i;
/* Native method getByte */
  sendByte(kComTerminalQuery);
  i = rxtime(1000);
  if (i == kComTerminalData) {
    i = rxtime(1000);
    return i;
  }
  return -1;
}
  
long isReset(void) {
/* test for reset */
  return reset;
}

void setReset(void) {
/* set reset */
  reset = true;
}

void clearReset(void) {
/* clear reset */
  reset = false;
}
  
#endif
      