#ifndef jvmException
#define jvmException

/*
****************************************
* File jvmException: exception handler *
****************************************
*/

/*
  Exception Handling.  
  Because exception handling is a major feature of the Java programming language, you might be surprised 
  to learn that exceptions are handled with a single bytecode instruction. When exceptions occur, the object 
  thrown must be a reference to an object of the subclass Throwable, and the current Java stackframe is 
  searched for the most recent catch clause that catches exceptions of this class or a superclass of this 
  class. When a matching catch clause is found, the program counter is reset to the address indicated by 
  the catch clause and execution continues from there. When no appropriate catch clause is found, that 
  frame is popped and the object is rethrown. If a catch clause is then found, the clause will contain the 
  location of the code for the exception, the program counter is reset to that location, and execution 
  continues. Otherwise, the frame is popped and the object is rethrown. 
  Note:
  A NullPointerException is thrown instead if the referenced object is null.  
*/

/*
  athrow
  Operation: Throw exception or error
  Format: athrow   
  Operand Stack: ...,objectref --> objectref

  Description
  The objectref must be of type reference and must refer to an object that is an instance of class Throwable
  or of a subclass of Throwable. It is popped from the operand stack. The objectref is then thrown by searching
  the current method for the first exception handler that matches the class of objectref.
  If an exception handler that matches objectref is found, it contains the location of the code intended to
  handle this exception. The pc register is reset to that location, the operand stack of the current frame is
  cleared, objectref is pushed back onto the operand stack, and execution continues. 
  If no matching exception handler is found in the current frame, that frame is popped. If the current frame
  represents an invocation of a synchronized method, the monitor acquired or reentered on invocation of the
  method is released or exited (respectively) as if by execution of a monitorexit instruction. Finally, the frame
  of its invoker is reinstated, if such a frame exists, and the objectref is rethrown. If no such frame exists,
  the current thread exits.
  Runtime Exceptions
  If objectref is null, athrow throws a NullPointerException instead of objectref.
  Notes
  The operand stack diagram for the athrow instruction may be misleading: If a handler for this exception is
  matched in the current method, the athrow instruction discards all the values on the operand stack, then pushes
  the thrown object onto the operand stack. However, if no handler is matched in the current method and the
  exception is thrown farther up the method invocation chain, then the operand stack of the method (if any) that
  handles the exception is cleared and objectref is pushed onto that empty operand stack. All intervening frames
  from the method that threw the exception up to, but not including, the method that handles the exception are
  discarded.
*/


//exceptions
#define  kIllegalArgumentException   1
#define  kNullPointerException       2
#define  kIndexOutOfBoundsException  3
#define  kOutOfMemoryError           4


#include "jvmStack.c"
#include "jvmData.c"
#include "jvmComm.c"
#include "jvmSubReturn.c"

  
long instanceOf(long instance, long class) {
  long instanceclass;
/* Auxiliary function (no bytecode) Meaning: Determine if an object is of given type and return result. */
/*
  Description: Determines whether the referenced object can be cast to be a reference to an 
  object of the expected class. It returns true or false.
*/
  if (instance != 0) {
    instanceclass = readJavaIntBE(instance); //get class type for instance
    do {
      if (instanceclass == class)
        return true;
      else
        instanceclass = readJavaIntBE(instanceclass); //get superclass of instance
    } while (instanceclass != 0xFFFF); //0xFFFF is superclass of Object class
  }
  return false;
}
  
void throw(long exceptionNumber) {
  long method;
/* Auxiliary function (no btecode): throw one of four runtime exceptions */
  method = readJavaIntBE(JVM_THROW);
  writeJavaIntBE(getJvmFP(),exceptionNumber); //write exceptionNumber to arg0 in current frame (eg. istore_0)
                                              //because method throwVMException starts with iload_0 
  setJvmEP(getJvmPC());                       //remember where exception is thrown
  setJvmPC(readJavaIntBE(method + JVM_CODE)); //set new pc to method throwVMException
}
  
void throwException(long e) {
  long exceptionTable,numExceptions,index,paramsize,localsize,pc,sp,class;
/* Auxiliary function (no bytecode): Throw an exception */
  if (e != 0) {
    if (getJvmEP() == 0) {
      setJvmEP(getJvmPC()); //set exception if not already set
    }
	do {
      exceptionTable = readJavaIntBE(getJvmMP() + JVM_EXCEPTION); //get address of exception table
      numExceptions = 0;
      if (exceptionTable != 0) {
        numExceptions = readJavaByte(exceptionTable); //read number of exception handlers
      }
	  index = 0;
      while (numExceptions > 0) {
        if ((getJvmPC() >= readJavaIntBE(exceptionTable + 1 + index)) && (getJvmPC() < readJavaIntBE(exceptionTable + 3 + index))) {
          if (instanceOf(e,readJavaIntBE(exceptionTable + 5 + index))) {
            //clear operand stack
            paramsize = (readJavaByte(getJvmMP() + JVM_PARAMETERS) & 0x7F); //mask out exception bit
            localsize = readJavaByte(getJvmMP() + JVM_LOCALS);
            sp = getJvmFP() + paramsize + localsize + 6;
            setJvmSP(sp);
            //push exception
            push(e);
            //jump to handler
            pc = readJavaIntBE(exceptionTable + 7 + index);
            setJvmPC(pc);
            return;
          }
		}
		index += 8; //size of exception record
        numExceptions--;
      }
	  jb_return();  //pop frame
    } while (getJvmMP() != 0xFFFF);
    class = readJavaIntBE(e);
    errException(class,getJvmEP()); //signal exception to IDE
    setJvmEP(0); //clear exception
  }
  else {
    throw(kNullPointerException);
  }  
}
      
void jb_athrow(void) {
  long aref,class;
/* Meaning: Throw exception or error. */
/*
  The objectref must be of type reference and must refer to an object that is an instance of class Throwable or
  of a subclass of Throwable. It is popped from the operand stack. The objectref is then thrown by searching the
  current method for the first exception handler that matches the class of objectref.
  If an exception handler that matches objectref is found, it contains the location of the code intended to handle
  this exception. The pc register is reset to that location, the operand stack of the current frame is cleared,
  objectref is pushed back onto the operand stack, and execution continues. 
  If no matching exception handler is found in the current frame, that frame is popped. If the current frame
  represents an invocation of a synchronized method, the monitor acquired or reentered on invocation of the method
  is released or exited (respectively) as if by execution of a monitorexit instruction.
  Finally, the frame of its invoker is reinstated, if such a frame exists, and the objectref is rethrown.
  If no such frame exists, the current thread exits.
*/
/*
  JEM format: athrow
*/
  aref = pop(); //get objectref
  throwException(aref);
}
    
#endif
    